/* ThreePartDate and BirthDate */

/* More room for month name text. */
.al_three_parts_date .col-year {
  padding-left: 0;
}
.al_three_parts_date .col-month {
  padding-right: 0;
}

/* Invalid feedback */
:root {
  --al-bs-exclamation: url("data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 12 12' width='12' height='12' fill='none' stroke='%23dc3545'%3e%3ccircle cx='6' cy='6' r='4.5'/%3e%3cpath stroke-linejoin='round' d='M5.8 3.6h.4L6 6.5z'/%3e%3ccircle cx='6' cy='8.2' r='.6' fill='%23dc3545' stroke='none'/%3e%3c/svg%3e");
}

.al_invalid {
  border: 1px solid var(--bs-danger);
  border-radius: 0.25rem;
  padding-right: calc(1.5em + 0.75rem);
  background-image: var(--al-bs-exclamation);
  background-repeat: no-repeat;
  background-position: right calc(0.375em + 0.1875rem) center;
  background-size: calc(0.75em + 0.375rem) calc(0.75em + 0.375rem);
}

.al_three_parts_date.al_invalid .col-year {
  padding-right: 0;
}

/* Avoid AL styles elongating date and year fields and codify widths. */
.al_three_parts_date .col > * {
  width: 100%;
}
